""" An example of using the Enthought class browser (enclbr) module. """

# Future Statements
from __future__ import print_function
# Standard library imports.
import sys, time

# Enthought library imports.
from envisage.developer.code_browser.api import CodeBrowser


def main(argv):
    """ Do it! """

    if len(argv) != 2:
        print('Usage: python example.py module_name')
        print
        print('e.g., python example.py traits')

    else:
        # Create a code browser.
        code_browser = CodeBrowser(filename='data.pickle')

        # Parse the specified package.
        start = time.time()
        contents = code_browser.read_package(sys.argv[1])
        stop = time.time()

        print('Time taken to parse', sys.argv[1], 'was', stop - start, 'secs')

        # Save the cache.
        code_browser.save('data.pickle')

    return

# Entry point for testing.
if __name__ == '__main__':
    main(sys.argv)

#### EOF ######################################################################






""" Provides a platform indepedent way to launch Editra. Code mostly copied
    from Editra run file (/usr/bin/editra) in Ubuntu wx package.
"""

import sys
import os

try:
    import Editra as Editra_root
    sys.path.insert(0, os.path.join(os.path.dirname(Editra_root.__file__),
                    'src'))
    # Really ugly: we need to remove the Editra name from the list of
    # imported modules
    sys.modules.pop('Editra')
except ImportError:
    import wx.tools

    # Editra needs its src package to be on the sys.path for plugins and
    # such to work right, so put it there before we do the first import of
    # any Editra package or module.
    sys.path.insert(0, os.path.join(os.path.dirname(wx.tools.__file__),
                                    'Editra', 'src'))

import Editra

# XXX: Super ugly: monkey-patch the main Editra class to add our plugins
old_MainLoop = Editra.Editra.MainLoop

def my_MainLoop(self, *args, **kwargs):
    from envisage.plugins.remote_editor.editor_plugins.editra.editra_plugin \
            import RemoteEditorPlugin
    plugin = RemoteEditorPlugin(Editra=self)
    try:
        plugin.do_PlugIt()
        old_MainLoop(self, *args, **kwargs)
    finally:
        plugin.client.unregister()

Editra.Editra.MainLoop = my_MainLoop

def main():
    Editra.Main()

if __name__ == '__main__':
    main()



#
#
# (C) Copyright 2011 Enthought, Inc., Austin, TX
# All right reserved.
#
# This file is open source software distributed according to the terms in LICENSE.txt
#
""" This module provides a plugin which adds an EventManager to application.

If the application does not already have an evt_mgr attribute which is an
instance of EventManager, the plugin creates a new EventManager instance,
creates a service to offer the event manager and sets the evt_mgr instance
of the application to the created event manager.
"""

# Enthought library imports.
from envisage.api import Plugin, ServiceOffer
from traits.api import List, Any


class EventManagerPlugin(Plugin):
    """ Plugin to add event manager to the application. """

    id = 'envisage.event_manager'

    SERVICE_OFFERS = 'envisage.service_offers'
    service_offers = List(contributes_to=SERVICE_OFFERS)

    def _service_offers_default(self):
        from encore.events.api import BaseEventManager, get_event_manager

        evt_mgr_service_offer = ServiceOffer(
            protocol   = BaseEventManager,
            factory    = get_event_manager,
        )
        return [evt_mgr_service_offer]




""" An action that configures an item in the project tree. """

# Enthought library imports.
from envisage.ui.single_project.project_action import ProjectAction

##############################################################################
# class 'ConfigureAction'
##############################################################################

class ConfigureAction(ProjectAction):
    """ An action that configures an item in the project tree.

    'Configures' in this sense means pop up a trait sheet!

    """

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """

        # fixme: We would like to use "kind='modal'" here, but it doesn't
        # work! The bug we see is that when the dialog is complete the
        # viscosity model assigned to the trait is NOT the same as the
        # viscosity model in the material's '_dirty_objects' collection!
        ui = event.node.obj.edit_traits(
            parent=event.window.control, kind='livemodal'
        )

        return

#### EOF ######################################################################



# Enthought library imports.
from pyface.action.api import Action, Group


class PreferencesAction(Action):
    """ An action that displays the preferences dialog.
    """

    #### 'Action' interface ###################################################

    # A longer description of the action.
    description = 'Open the preferences dialog'

    # The action's name (displayed on menus/tool bar tools etc).
    name = 'Prefere&nces...'

    # A short description of the action used for tooltip text etc.
    tooltip = 'Open the preferences dialog'

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        from envisage.ui.tasks.preferences_dialog import \
            PreferencesDialog

        window = event.task.window
        dialog = window.application.get_service(PreferencesDialog)
        ui = dialog.edit_traits(parent=window.control, kind='livemodal')

        if ui.result:
            window.application.preferences.save()


class PreferencesGroup(Group):
    """ A group that contains the preferences action.
    """

    #### 'Action' interface ###################################################

    # The group's identifier (unique within action manager).
    id = 'PreferencesGroup'

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        super(PreferencesGroup, self).__init__(PreferencesAction(), **traits)



""" A view showing a summary of the running application. """


# Standard library imports.
import inspect

# Enthought library imports.
from envisage.api import IApplication, IExtensionPoint
from envisage.api import IExtensionRegistry
from envisage.developer.code_browser.api import CodeBrowser
from apptools.io.api import File
from traits.api import HasTraits, Instance
from traitsui.api import Item, TreeEditor, View

# fixme: non-api import.
from envisage.plugins.text_editor.editor.text_editor import TextEditor

# Local imports.
from .extension_registry_browser_tree import \
     extension_registry_browser_tree_nodes


extension_registry_browser_view = View(
    Item(
        name       = 'extension_registry',
        show_label = False,
        editor     = TreeEditor(
            nodes       = extension_registry_browser_tree_nodes,
            editable    = False,
            orientation = 'vertical',
            hide_root   = True,
            show_icons  = True,
            on_dclick   = 'object.dclick'
        )
    ),

    resizable = True,
    style     = 'custom',
    title     = 'Extension Registry',

    width     = .1,
    height    = .1
)


class ExtensionRegistryBrowser(HasTraits):
    """ An extension registry browser.

    Actually, this class exists just because to use a trait editor we have
    to have a trait to edit!

    """

    #### 'ExtensionRegistryBrowser' interface #################################

    # The application that whose extension registry we are browsing.
    application = Instance(IApplication)

    # The code browser that we use to parse plugin source code.
    code_browser = Instance(CodeBrowser)

    # The extension registry that we are browsing.
    extension_registry = Instance(IExtensionRegistry)

    # The workbench service.
    workbench = Instance('envisage.ui.workbench.api.Workbench')

    # The default traits UI view.
    traits_view = extension_registry_browser_view

    ###########################################################################
    # 'ExtensionRegistryBrowser' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _extension_registry_default(self):
        """ Trait initializer. """

        return self.application

    def _workbench_default(self):
        """ Trait initializer. """

        from envisage.ui.workbench.api import Workbench

        return self.application.get_service(Workbench)

    #### Methods ##############################################################

    def dclick(self, obj):
        """ Called when an object in the tree is double-clicked. """

        # Double-click on an extension point.
        if IExtensionPoint(obj, None) is not None:
            self.dclick_extension_point(obj)

        # Double-click on an extension.
        elif IExtensionPoint(obj.parent.value, None) is not None:
            self.dclick_extension(obj)

        return

    def dclick_extension_point(self, obj):
        """ Called when an extension point is double-clicked. """

        # Find the plugin that offered the extension point.
        plugin = self._get_plugin(obj)

        # Parse the plugin source code.
        module = self._parse_plugin(plugin)

        # Get the plugin klass.
        klass = self._get_plugin_klass(module, plugin)

        # Edit the plugin.
        editor = self.workbench.edit(
            self._get_file_object(plugin), kind=TextEditor
        )

        # Was the extension point offered declaratively via a trait?
        trait_name = self._get_extension_point_trait(plugin, obj.id)
        if trait_name is not None:
            attribute = klass.attributes.get(trait_name)
            lineno    = attribute.lineno

        else:
            lineno = klass.lineno

        editor.select_line(lineno)

        return

    def dclick_extension(self, obj):
        """ Called when an extension is double-clicked. """

        extension_point = obj.parent.value
        index           = obj.parent._index

        extension_registry = self.extension_registry
        extensions = extension_registry.extension_registry._extensions

        total = 0
        provider_index = 0
        for l in extensions[extension_point.id]:
            total = total + len(l)
            if index < total:
                break
            provider_index += 1


        plugin = extension_registry.extension_registry._providers[provider_index]

        # Parse the plugin source code.
        module = self._parse_plugin(plugin)

        # Get the plugin klass.
        klass = self._get_plugin_klass(module, plugin)

        # Edit the plugin.
        editor = self.workbench.edit(
            self._get_file_object(plugin), kind=TextEditor
        )

        # Was the extension offered declaratively?
        trait_name = self._get_extension_trait(plugin, extension_point.id)
        if trait_name is not None:

            # Does the trait have a default initializer?
            initializer = klass.methods.get('_%s_default' % trait_name)
            if initializer is not None:
                lineno    = initializer.lineno

            else:
                attribute = klass.attributes.get(trait_name)
                lineno    = attribute.lineno

        else:
            lineno = klass.lineno

        editor.select_line(lineno)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_extension_trait(self, plugin, id):
        """ Return the extension trait with the specifed Id.

        Return None if the extension point was not declared via a trait.

        """

        extension_traits = plugin.traits(contributes_to=id)

        if len(extension_traits) > 0:
            # There is *at most* one extension point trait per extension point.
            trait_name = next(iter(extension_traits))

        else:
            trait_name = None

        return trait_name

    def _get_extension_point_trait(self, plugin, id):
        """ Return the extension point trait with the specifed Id.

        Return None if the extension point was not declared via a trait.

        """

        extension_point_traits = plugin.traits(__extension_point__=True)

        for trait_name, trait in extension_point_traits.items():
            if trait.trait_type.id == id:
                break

        else:
            trait_name = None

        return trait_name

    def _get_plugin(self, extension_point):
        """ Return the plugin that offered an extension point. """

        for plugin in self.application:
            if extension_point in plugin.get_extension_points():
                break

        else:
            plugin = None

        return plugin

    def _get_plugin_klass(self, module, plugin):
        """ Get the klass that defines the plugin. """

        for name, klass in module.klasses.items():
            if name == type(plugin).__name__:
                break

        else:
            klass = None

        return klass

    def _get_file_object(self, obj):
        """ Return a 'File' object for the object's source file. """

        return File(path=inspect.getsourcefile(type(obj)))

    def _parse_plugin(self, plugin):
        """ Parse the plugin source code. """

        filename = self._get_file_object(plugin).path

        return self.code_browser.read_file(filename)

#### EOF ######################################################################



""" An action that displays the preferences dialog. """


# Enthought library imports.
from pyface.api import ImageResource
from pyface.action.api import Action


class EditPreferencesAction(Action):
    """ An action that displays the preferences dialog. """

    #### 'Action' interface ###################################################

    # A longer description of the action.
    description = 'Manage Preferences'

    # The action's image (displayed on tool bar tools etc).
    image = ImageResource('preferences')

    # The action's name (displayed on menus/tool bar tools etc).
    name = 'Preferences'

    # A short description of the action used for tooltip text etc.
    tooltip = 'Manage Preferences'

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """

        from apptools.preferences.ui.api import PreferencesManager

        # Lookup the preferences manager service.
        manager = event.window.application.get_service(PreferencesManager)
        ui = manager.edit_traits(parent=event.window.control, kind='modal')

        # If the user hit the "Ok" button, then save the preferences in case
        # application crashes before it exits!
        if ui.result:
            self.window.application.preferences.save()

        return

#### EOF ######################################################################



""" The resource manager interface. """


# Enthought library imports.
from traits.api import Instance, Interface

# Local imports.
from .i_resource_protocol import IResourceProtocol


class IResourceManager(Interface):
    """ The resource manager interface. """

    # The protocols used by the manager to resolve resource URLs.
    resource_protocols = Instance(IResourceProtocol)

    def file(self, url):
        """ Return a readable file-like object for the specified url.

        Raise a 'NoSuchResourceError' if the resource does not exist.

        e.g.::

          manager.file('pkgfile://acme.ui.workbench/preferences.ini')

        """

#### EOF ######################################################################



""" The interface for extension providers. """


# Enthought library imports.
from traits.api import Event, Interface

# Local imports.
from .extension_point_changed_event import ExtensionPointChangedEvent


class IExtensionProvider(Interface):
    """ The interface for extension providers. """

    # The event fired when one of the provider's extension points has changed.
    extension_point_changed = Event(ExtensionPointChangedEvent)

    def get_extension_points(self):
        """ Return the extension points offered by the provider.

        Return an empty list if the provider does not offer any extension
        points.

        """

    def get_extensions(self, extension_point_id):
        """ Return the provider's extensions to an extension point.

        The return value *must* be a list. Return an empty list if the provider
        does not contribute any extensions to the extension point.

        """

#### EOF ######################################################################



""" An implementation of weak references that works for bound methods.

This code is based on the code in the Python Cookbook, but you can call `ref`
for objects that are *not* bound methods too, in which case it just returns a
standard `weakref.ref`.

Weak references to bound methods are cached so that `ref(x) is ref(x)` as for
standard weakrefs, and the `ref` class defined here is therefore intended to be
used as a drop-in replacement for 'weakref.ref'.

"""

# Standard library imports.
import sys, weakref

# Because this module is intended as a drop-in replacement for weakref, we
# import everything from that module here (so the user can do things like
# "import safeweakref as weakref" etc).

from weakref import *

__all__ = weakref.__all__

if hasattr(weakref, "WeakMethod"):
    pass
else:
    # Backport WeakMethod from Python 3
    class WeakMethod(weakref.ref):
        """
        A custom `weakref.ref` subclass which simulates a weak reference to
        a bound method, working around the lifetime problem of bound methods.
        """

        __slots__ = "_func_ref", "_meth_type", "_alive", "__weakref__"

        def __new__(cls, meth, callback=None):
            try:
                obj = meth.__self__
                func = meth.__func__
            except AttributeError:
                raise TypeError(
                    "argument should be a bound method, not {}"
                    .format(type(meth))
                )
            def _cb(arg):
                # The self-weakref trick is needed to avoid creating a reference
                # cycle.
                self = self_wr()
                if self._alive:
                    self._alive = False
                    if callback is not None:
                        callback(self)
            self = weakref.ref.__new__(cls, obj, _cb)
            self._func_ref = weakref.ref(func, _cb)
            self._meth_type = type(meth)
            self._alive = True
            self_wr = weakref.ref(self)
            return self

        def __call__(self):
            obj = weakref.ref.__call__(self)
            func = self._func_ref()
            if obj is None or func is None:
                return None
            return self._meth_type(func, obj)

        def __eq__(self, other):
            if isinstance(other, WeakMethod):
                if not self._alive or not other._alive:
                    return self is other
                return weakref.ref.__eq__(self, other) and self._func_ref == other._func_ref
            return False

        def __ne__(self, other):
            if isinstance(other, WeakMethod):
                if not self._alive or not other._alive:
                    return self is not other
                return weakref.ref.__ne__(self, other) or self._func_ref != other._func_ref
            return True

        __hash__ = weakref.ref.__hash__


class ref(object):
    """ An implementation of weak references that works for bound methods and \
    caches them. """
    _cache = weakref.WeakKeyDictionary()

    def __new__(cls, obj, callback=None):
        if getattr(obj, "__self__", None) is not None:  # Bound method
            # Caching
            func_cache = cls._cache.setdefault(obj.__self__, {})
            self = func_cache.get(obj.__func__)
            if self is None:
                self = WeakMethod(obj, callback)
                func_cache[obj.__func__] = self
            return self
        else:
            return weakref.ref(obj, callback)

#### EOF ######################################################################



""" A Python namespace. """


# Standard library imports.
import os, sys

# Enthought library imports.
from traits.api import Dict, HasTraits


class Namespace(HasTraits):
    """ A Python namespace. """

    #### 'Namespace' interface ################################################

    # All of the names defined within the module.
    locals = Dict

    # All of the names imported into the module.
    imports = Dict

    # fixme: This is a dict from name -> bool indicating whether a name in
    # locals is a trait or not.  Should locals be a dict containing a tuple?
    _is_trait = Dict

    ###########################################################################
    # 'Namespace' interface.
    ###########################################################################

    def is_trait(self, name):
        """ Attempt to resolve a name to see if it is a trait. """

        # fixme: We might want to be a bit cleverer than this ;^)
        return True

        # Try the namespace's locals first.
        if name in self.locals:
            # These are the definitive traits markers!
            if name in ['TraitFactory', 'Trait', 'Instance']:
                is_trait = True

            # The name won't be in the '_is_trait' dictionary if it was not
            # defined by an assignment i.e., it is a class or function!
            elif not self._is_trait.get(name, False):
                is_trait = False

            else:
                next = self.locals[name]
                if len(next.source) > 0:
                    is_trait = self.is_trait(next.source)

                else:
                    is_trait = False

        # Is the name imported?
        elif self.is_imported(name):
            module_name = self.get_next_module_name(name)
            module = self.get_next_module(module_name)

            if module is not None:
                components = name.split('.')
                is_trait = module.is_trait(components[-1])

            else:
                is_trait = False

        # If we have a parent namespace then try that.
        elif self.namespace is not None:
            is_trait = self.namespace.is_trait(name)

        # Otherwise give up!
        else:
            is_trait = False

        return is_trait

    def is_imported(self, name):
        """ Returns TRUE if a name is imported. """

        components = name.split('.')

        return components[0] in self.imports

    def get_next_module_name(self, symbol):
        """ Returns the name of the module that a symbol was imported from. """

        components = symbol.split('.')
        if len(components) == 1:
            module_name = self.imports[symbol]

        else:
            path = []

            # The first component MUST have been imported.
            #
            # An empty string means 'import' as opposed to 'from' used.
            value = self.imports[components[0]]
            if len(value) > 0:
                path.append(value)

                module_name = '.'.join(path)

                return module_name

            path.append(components[0])

            for component in components[1:-1]:
                path.append(component)

            module_name = '.'.join(path)

        return module_name

    def get_next_module(self, module_name):
        """ Returns a parsed module with the specified name.

        Returns None if the module cannot be found or is 'ignored'.  We ignore
        all modules in the Python core, extension modules etc.

        """

        # fixme: Circular imports!
        from .enclbr import find_module, read_file

        # Try to find the module that the symbol came from.
        dirname = os.path.dirname(self.filename)
        filename = find_module(module_name, [dirname] + sys.path)

        if filename is not None:
            # If the filename refers to a directory then it must be a
            # package.
            if os.path.isdir(filename):
                filename = os.path.join(filename, '__init__.py')

            # fixme: We should probably put in some more formal filter
            # mechanism here!
            #
            # Ignore extension modules.
            if filename.endswith('.pyd') or filename.endswith('.so'):
                module = None

            # Ignore standard modules.
            elif filename.startswith(sys.prefix):
                module = None

            # Parse it!
            else:
                module = read_file(filename)

        else:
            module = None

        return module

#### EOF ######################################################################



""" A hook to allow code be executed when a class is loaded. """


# Standard library imports.
import sys

# Enthought library imports.
from traits.api import Callable, HasTraits, MetaHasTraits, Str


class ClassLoadHook(HasTraits):
    """ A hook to allow code to be executed when a class is loaded.

    If the class is *already* loaded when the 'connect' method is called then
    the code is executed immediately.

    """

    #### 'ClassLoadHook' interface ############################################

    # The name of the class. When this class is loaded the 'on_class_loaded'
    # method is called.
    class_name = Str

    # A callable that will be executed when the class is loaded. The callable
    # must take a single argument which will be the loaded class.
    #
    # This is used in the default implementation of 'on_class_loaded'. If you
    # override that, then you don't have to set to this trait.
    on_load = Callable

    ###########################################################################
    # 'ClassLoadHook' interface.
    ###########################################################################

    def connect(self):
        """ Connect the load hook to listen for the class being loaded. """

        MetaHasTraits.add_listener(self.on_class_loaded, self.class_name)

        # If the class has already been loaded then run the code now!
        cls = self._get_class(self.class_name)
        if cls is not None:
            self.on_class_loaded(cls)

        return

    def disconnect(self):
        """ Disconnect the load hook. """

        MetaHasTraits.remove_listener(self.on_class_loaded, self.class_name)

        return

    def on_class_loaded(self, cls):
        """ This method is called when the class is loaded.

        If 'self.on_load' is not None, it calls 'self.on_load(cls)'.

        """

        if self.on_load is not None:
            self.on_load(cls)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_class(self, class_path):
        """ Returns the class defined by *class_path*.

        Returns **None** if the class has not yet been loaded.

        """

        # Only check if the class name has at least a partial hierarchy.
        #
        # fixme: Comment should say why!
        if '.' in class_path:
            components = class_path.split('.')

            module_name = '.'.join(components[:-1])
            class_name  = components[-1]

            # The class is loaded if its module has been imported and the class
            # is defined in the module dictionary.
            module = sys.modules.get(module_name, None)
            if module is not None and hasattr(module, class_name):
                klass = getattr(module, class_name)

            else:
                klass = None

        else:
            klass = None

        return klass

#### EOF ######################################################################



""" Classes used to represent modules. """


# Standard library imports.
import ast, os

# Enthought library imports.
from apptools.io.api import File
from traits.api import Any, Dict, HasTraits, Instance, Str

# Local imports.
from .assign import Assign, AssignFactory
from .function import Function, FunctionFactory
from .klass import Klass, KlassFactory
from .namespace import Namespace


class Module(Namespace):
    """ A module. """

    #### 'Module' interface ###################################################

    # The namespace that the module is defined in. If the module is in a
    # package then this will be the appropriate 'Package' instance, otherwise
    # it will be None.
    namespace = Instance(Namespace)

    # The absolute filename of the module.
    filename = Str

    # The name of the module (this is currently the fully qualified name i.e.,
    # it includes the names of the packages that the module is contained in).
    name = Str

    # The module's doc-string (None if there is no doc string, a string if
    # there is).
    doc = Any

    # The module-level attributes.
    attributes = Dict

    # The classes defined in the module.
    klasses = Dict

    # The module-level functions.
    functions = Dict

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __str__(self):
        """ Returns an informal string representation of the object. """

        return 'Module %s' % self.filename

    def __getstate__(self):
        """ Returns the state of the object for pickling. """

        state = {}
        state['namespace'] = self.namespace
        state['filename'] = self.filename
        state['name'] = self.name
        state['doc'] = self.doc
        state['attributes'] = self.attributes
        state['klasses'] = self.klasses
        state['functions'] = self.functions

        return state

    def __setstate__(self, state):
        """ Sets the state object duting unpickling. """

        self.namespace = state['namespace']
        self.filename = state['filename']
        self.name = state['name']
        self.doc = state['doc']
        self.attributes = state['attributes']
        self.klasses = state['klasses']
        self.functions = state['functions']

        return


class ModuleFactory(HasTraits):
    """ A factory for modules. """

    ###########################################################################
    # 'ModuleFactory' interface.
    ###########################################################################

    def from_file(self, filename, namespace=None):
        """ Creates a module by parsing a file. """

        # Parse the file.
        with open(filename) as f:
            node = ast.parse(f.read())

        # Create a new module.
        module = Module(
            namespace = namespace,
            filename  = filename,
            name      = self._get_module_name(filename),
            doc       = ast.get_docstring(node, clean=False),
        )

        # Walk the AST picking out the things we care about!
        ModuleVisitor(module).visit(node)

        return module

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_module_name(self, filename):
        """ Get the fully qualified module name for a filename.

        e.g. if the filename is:-

        '/enthought/envisage/core/core_plugin_definition.py'

        we would return:-

        'envisage.core.core_plugin_definition'

        """

        # Get the name of the module minus the '.py'
        module, ext = os.path.splitext(os.path.basename(filename))

        # Start with the actual module name.
        module_path = [module]

        # If the directory is a Python package then add it to the module path.
        parent = File(os.path.dirname(filename))
        while parent.is_package:
            module_path.insert(0, parent.name)
            parent = parent.parent

        return '.'.join(module_path)


class ModuleVisitor(ast.NodeVisitor):
    """ An AST visitor for top-level modules. """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, module):
        """ Creates a new visitor. """

        self.module = module

        # Factories used to create klasses, functions and assignments from
        # AST nodes.
        self._klass_factory    = KlassFactory()
        self._function_factory = FunctionFactory()
        self._assign_factory   = AssignFactory()

        return

    ###########################################################################
    # 'NodeVisitor' interface.
    ###########################################################################

    def visit_ClassDef(self, node):
        """ Visits a class node. """
        klass = self._klass_factory.from_ast(self.module, node)

        self.module.locals[node.name] = klass
        self.module.klasses[node.name] = klass

        return

    def visit_FunctionDef(self, node):
        """ Visits a function node. """

        function = self._function_factory.from_ast(self.module, node)

        self.module.locals[node.name] = function
        self.module.functions[node.name] = function

        return

    def visit_Assign(self, node):
        """ Visits an assignment node. """

        assign = self._assign_factory.from_ast(self.module, node)

        # Does the assigment look like it *might* be a trait? (i.e., it is NOT
        # an expression or a constant etc.).
        if len(assign.source) > 0:
            assign.is_trait = self.module.is_trait(assign.source)

        else:
            assign.is_trait = False

        for target in assign.targets:
            self.module.locals[target] = assign
            self.module.attributes[target] = assign
            self.module._is_trait[target] = assign.is_trait

        return

    def visit_ImportFrom(self, node):
        """ Visits a from node. """

        for alias in node.names:
            self.module.imports[alias.name] = node.module

        return

    def visit_Import(self, node):
        """ Visits an import node. """

        for alias in node.names:
            # fixme: We currently use the fact that we add an empty string to
            # the imports dictionary to tell the difference later on between
            # 'import' and 'from import'.
            self.module.imports[alias.name] = ''

        return

#### EOF ######################################################################




# Enthought library imports
from envisage.api import Plugin
from traits.api import Bool, Callable, Event, Int, Instance, Str, Trait

# Local, relative imports
from .update_info import UpdateInfo


# The globally unique ID of this plugin
ID = "envisage.plugins.update_checker"


class UpdateCheckerPlugin(Plugin):

    name = "Update Checker Plugin"

    # The URI to an updates.xml file.  Although this can be a local path,
    # typically it will be a URL.
    location = Str("localhost")

    # Update

    # Should the plugin automatically check for updates?
    check_automatically = Bool(True)

    # The frequency with which to check for updates
    # Can be "startup", which means at application startup, or an integer number
    # of seconds
    # TODO: Contribute a preference for this!
    check_frequency = Trait("startup", "startup", Int)

    # Whether or not to display a dialog informing the user of available
    # updates
    display_dialog = Bool(True)

    # The UpdateInfo object that contains the actual update information.
    # Regardless of whether or not a dialog is displayed to the user, this
    # attribute will be populated with the appropriate information.
    update_info = Instance(UpdateInfo)

    # This event fires every time the UpdateCheckerPlugin is invoked and
    # determines that an application update is available.
    update_needed = Event()  #Event(ApplicationEvent)

    #========================================================================
    # Plugin interface
    #========================================================================

    def start(self):
        """ Make the appropriate contributions
        """

        self.application.on_trait_change("started", self._check_for_update)

    def stop(self):
        """ Clean up
        """

        self.application.on_trait_change("started", self._check_for_update, remove=True)

    #========================================================================
    # Public methods
    #========================================================================

    def _check_for_update(self):
        pass





from envisage.ui.action.api import Action, ActionSet, Group
from pyface.api import FileDialog, OK
from pyface.action.api import Action as PyfaceAction
from envisage.plugins.remote_editor.api import IRemoteEditor



def get_server(window):
    """ Given an application window, retrieve the communication server.
    """
    return window.application.get_service(IRemoteEditor)

################################################################################
# Groups
################################################################################
file_group = Group(
    id='RemoteEditorFileGroup',
    path='MenuBar/File',
    before='ExitGroup'
)


################################################################################
# `OpenScript` class.
################################################################################
class OpenScript(PyfaceAction):
    """ An action that opens a Python file in a remote editor. """

    tooltip      = "Open a Python script in separate editor."

    description  = "Open a Python script in separate editor."


    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """
        server = get_server(self.window)

        wildcard = 'Python files (*.py)|*.py'

        parent = self.window.control
        dialog = FileDialog(parent=parent,
                            title='Open Python script in separate editor',
                            action='open', wildcard=wildcard
                            )
        if dialog.open() == OK:
            server.open_file(dialog.path)


open_script = Action(
    path        = "MenuBar/File",
    class_name  = __name__ + '.OpenScript',
    name        = "Open script in editor",
    group       = "RemoteEditorFileGroup",
)

################################################################################
# `NewScript` class.
################################################################################
class NewScript(PyfaceAction):
    """ An action that opens a new file in a remote editor. """

    tooltip       = "Open a new file in separate editor."

    description   = "Open a new file in separate editor."

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """
        server = get_server(self.window)
        server.new_file()


new_script = Action(
    path        = "MenuBar/File",
    class_name  = __name__ + '.NewScript',
    name        = "New script in editor",
    group       = "RemoteEditorFileGroup",
)


################################################################################
# `RemoteEditorActionSet` class.
################################################################################
class RemoteEditorActionSet(ActionSet):
    """ The default action set for the remote editor plugin. """

    groups = [file_group, ]

    actions = [open_script, new_script]



from .new_project_action import NewProjectAction
from .open_project_action import OpenProjectAction
from .save_project_action import SaveProjectAction
from .save_as_project_action import SaveAsProjectAction
from .close_project_action import CloseProjectAction
from .switch_to_action import SwitchToAction



# This file was automatically generated from the `setup.py` script.
version = '4.6.0'
full_version = '4.6.0'
git_revision = '40fc473'
is_released = True

if not is_released:
    version = full_version



""" An extension registry that uses plugins as extension providers. """


# Enthought library imports.
from traits.api import Instance, on_trait_change

# Local imports.
from .i_plugin_manager import IPluginManager
from .provider_extension_registry import ProviderExtensionRegistry


class PluginExtensionRegistry(ProviderExtensionRegistry):
    """ An extension registry that uses plugins as extension providers.

    The application's plugins are used as the registries providers so adding
    or removing a plugin affects the extension points and extensions etc.

    """

    #### 'PluginExtensionRegistry' interface ##################################

    # The plugin manager that has the plugins we are after!
    plugin_manager = Instance(IPluginManager)

    ###########################################################################
    # 'PluginExtensionRegistry' interface.
    ###########################################################################

    #### Trait change handlers ################################################

    def _plugin_manager_changed(self, trait_name, old, new):
        """ Static trait change handler. """

        # In practise I can't see why you would ever want (or need) to change
        # the registry's plugin manager on the fly, but hey... Hence, 'old'
        # will probably always be 'None'!
        if old is not None:
            for plugin in old:
                self.remove_provider(plugin)

        if new is not None:
            for plugin in new:
                self.add_provider(plugin)

        return

    @on_trait_change('plugin_manager:plugin_added')
    def _on_plugin_added(self, obj, trait_name, old, event):
        """ Dynamic trait change handler. """

        self.add_provider(event.plugin)

        return

    @on_trait_change('plugin_manager:plugin_removed')
    def _on_plugin_removed(self, obj, trait_name, old, event):
        """ Dynamic trait change handler. """

        self.remove_provider(event.plugin)

        return

#### EOF ######################################################################



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2007 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

"""
A base class for projects that can be displayed by the single_project
plugin.

"""

# Standard library imports.
import logging
import os
import sys

# Enthought library imports
from traits.etsconfig.api import ETSConfig
import apptools.sweet_pickle
from apptools.io.api import File
from traits.api import Any, Bool, Dict, Directory, HasTraits, \
    Instance, Property, Str
from traitsui.api import Group, View
from traits.util.clean_strings import clean_filename

# Local imports.
#from envisage.ui.single_project.editor.project_editor import \
#    ProjectEditor
from envisage.api import Application


# Setup a logger for this module.
logger = logging.getLogger(__name__)


class Project(HasTraits):
    """
    A base class for projects that can be displayed by the single_project
    plugin.

    """

    ##########################################################################
    # CLASS Attributes
    ##########################################################################

    #### public 'Project' class interface ####################################

    # Indicates whether instances of this project class are stored as files or
    # directories.  The rest of the single_project plugin will follow this
    # setting when using this project class.
    #
    # This is meant to be a constant for the lifetime of this class!
    PROJECTS_ARE_FILES = True

    # Current envisage application.
    application = Instance(Application, transient=True)


    #### protected 'Project' class interface #################################

    # Format used to create a unique name from a location and a counter.
    _unique_name_format = '%s_%s'


    ##########################################################################
    # Attributes
    ##########################################################################

    #### public 'Project' interface ##########################################

    # True if this project contains un-saved state (has been modified.)
    dirty = Bool(False, transient=True)

    # True if we allow save requests on the project.
    is_save_allowed = Bool(True, transient=True)

    # True if we allow save_as requests on the project.
    is_save_as_allowed = Bool(True, transient=True)

    # The location of this project on the filesystem.  The default value
    # depends on the runtime environment so we use a traits default method to
    # set it.
    location = Directory

    # The name of this project.  This is calculated from the project's current
    # location or, if there is no location, from a default value.  See the
    # property's getter method.
    name = Property(Str)

    # The UI view to use when creating a new project
    traits_view = View(
        Group('location'),
        title = 'New Project',
        id = 'envisage.single_project.project.Project',
        buttons = [ 'OK', 'Cancel' ],
        width = 0.33,

        # Ensure closing via the dialog close button is the same
        # as clicking cancel.
        close_result = False,

        # Ensure the user can resize the dialog.
        resizable = True,
        )


    #### protected 'Project' interface #######################################

    # A list of editors currently open to visualize our resources
    # FIXME: Re-add the ProjectEditor's(if we need them) once they are fixed.
    #_editors = Dict(Any, ProjectEditor, transient=True)

    # The cache of this project's name.  We can't just initialize it to a
    # default value since the location may be cleared at any time.
    _name = Str(transient=True)


    ##########################################################################
    # 'Object' interface.
    ##########################################################################

    #### operator methods ####################################################

    def __getstate__(self):
        """
        Get the state of this object for pickling.

        Extended to limit which attributes get pickled and also to add
        version numbers to our pickle.

        It is STRONGLY recommended that derived classes not leverage the
        pickling mechanism if they wish to store additional data in OTHER
        files as part of their project.  Instead, they should override the
        *_save* method.

        Extending this method in derived classes IS appropriate if you want
        to store additional data in the SAME pickle file as the project.

        """

        # Determine the starting point for our state.
        state = super(Project, self).__getstate__().copy()

        # Remove any transient traits.
        for trait_name in self.trait_names(transient=True):
            state.pop(trait_name, None)

        # Add in our current version number.  Note use a different attribute
        # name from any base or derived class so that our numbers don't
        # override theirs.
        state['_project_version_major'] = 1
        state['_project_version_minor'] = 0

        return state


    def __setstate__(self, state):
        """
        Restore the state of this object during unpickling.

        Extended to upgrade pickles to the current Project version.

        It is STRONGLY recommended that derived classes not leverage the
        unpickling mechanism if they wish to load data from additional pickle
        files. Instead, they should override the *_load* method.

        Extending this method in derived classes IS appropriate if you want
        to load additional data from the SAME pickle file as the project.

        """

        # Get the version info out of the state dictionary.
        major = state.pop('_project_version_major', 0)
        minor = state.pop('_project_version_minor', 0)

        # Upgrade to version 1.
        if major < 1:

            # Remove any old attributes.
            # - name is now a calculated property instead of stored value.
            for key in ['name']:
                state.pop(key, None)

        # Restore our state.
        return super(Project, self).__setstate__(state)


    def __str__(self):
        """
        Return the unofficial string representation of this object.

        """

        result ='%s(name=%s)' % (super(Project, self).__str__(), self.name)
        return result


    ##########################################################################
    # 'Project' interface.
    ##########################################################################

    #### public interface ####################################################

    def get_default_project_location(self, application):
        """
        Return the default location for a new project.

        """

        path = self.get_default_path(application)
        name = clean_filename(self.get_default_name())
        location = os.path.join(path, name)
        location = self._make_location_unique(location)

        return location

    def get_default_path(cls, application):
        """
        Return the default path to the parent directory for a new project.

        """

        # When unpickling, we don't have a reference to the current application, so we
        # fallback on application_home.
        if application is None:
            return ETSConfig.application_home

        app_preferences = application.preferences
        path_id = 'envisage.ui.' \
            'single_project.preferred_path'
        path = app_preferences.get(path_id)

        # If the 'preferred_path' variable isn't set in the user's preferences,
        # then we set to the the application home by default.
        if len(path) == 0:
            app_home = ETSConfig.application_home
            app_preferences.set(path_id, app_home)
            return app_home

        return path

    get_default_path = classmethod(get_default_path)

    def get_default_name(self):
        """
        Return the default name for a new project.

        """

        return 'New Project'

    def get_pickle_filename(cls, location):
        """
        Generate the project's pickle filename given a source location.

        By default, the filename IS the specified location or, when saving
        projects as directories, a file called 'project' within a directory
        that is the specified location,

        Derived classes may wish to use the location as the basis for
        identifying the real pickle file.

        """

        if cls.PROJECTS_ARE_FILES:
            result = location
        else:
            result = os.path.join(location, 'project')

        return result

    get_pickle_filename = classmethod(get_pickle_filename)

    def get_pickle_package(cls):
        """
        Returns the pickle package to use for pickling and unpickling
        projects.

        Implementors can override this to customize the way in which
        projects are pickled and unpickled.

        This implementation returns the apptools.sweet_pickle package which
        supports versioning and refactoring of classes.

        """

        return apptools.sweet_pickle

    get_pickle_package = classmethod(get_pickle_package)

    def load(cls, location, application):
        """
        Load a project from a specified location.

        The loaded project's location is always set to the location the project
        was actually loaded from.  Additionally, the dirty flag is cleared
        on the loaded project.

        An exception will be raised to indicate a failure.

        """

        # Load the project in a manner that derived classes can modify.
        project = cls._load(location)

        # Ensure the project's location reflects where the project was loaded
        # from and that the dirty flag is not set.
        project.location = location
        project.dirty = False

        # Set the project's 'application' to the running application passed in.
        project.application = application

        return project

    load = classmethod(load)

    def register_editor(self, resource, editor, remove=False):
        """
        Inform us that an editor has been opened for what is believed to be
        a resource in this project.

        Note that if this project can be represented as a hierarchy, the
        resource may not be a top level object in that hierarchy!

        """

        # Warn if the resource is not part of this project
        if not self._contains_resource(resource):
            logger.warn('This Project [%s] does not contain resource [%s]' % \
                (self, resource))

        # Add or remove from our set of editors as requested
        if not remove:
            self._editors[resource] = editor
        else:
            del self._editors[resource]


    def save(self, location=None, overwrite=False):
        """
        Save this project.

        The project is saved to its current location, identified by the value
        of the *location* trait, unless a new location is explicitly provided.
        The specification of a new location is used to do a 'save as'
        operation.

        If a new location is provided, and a file or directory already exists
        at that location, then an *AssertionError* exception is raised unless
        the overwrite flag is True.  This ensures that users won't accidentally
        overwrite existing data.

        This method requires the overwrite flag because prompting the user to
        confirm the overwrite requires GUI interaction and thus should not be
        done at the model level.

        Note that, because we can rely on the location of a loaded project
        always being set to the location it was loaded from, there is no reason
        to try to force the *location* trait within a saved project to the
        location we are trying to save to.  Instead, we update the value in
        the in-memory version of the project only if the save completed
        successfully.

        The dirty flag is always cleared upon a succesfull save of the project.

        An exception will be raised to indicate a failure.

        """

        # Ensure saving (or save as) is allowed at this time.
        if location is None or location == self.location:
            if not self.is_save_allowed:
                raise AssertionError('Saving is currently not allowed.')
        elif location != self.location:
            if not self.is_save_as_allowed:
                raise AssertionError('Save as is currently not allowed.')

        # Use the internally-specified location unless a new location was
        # explicitly provided.  The new location can not contain any starting
        # or trailing whitespace and it cannot overwrite an existing file or
        # directory unless that was explicitly allowed.
        loc = self.location
        if location is not None:
            location = location.strip()
            if len(location) > 0 and location != self.location:

                # Ensure we never overwrite existing files / directories just
                # because someone specified a new location.  (Confirmation or
                # correction of overwriting requires prompting of the user and
                # is thus not part of the project model.)
                if os.path.exists(location) and overwrite is False:
                    raise AssertionError('Can not overwrite existing ' + \
                        'location [%s]' % location)

                # The requested location is valid so let's use it.
                loc = location

        # Ensure all necessary directories exist.  If we're saving a file, then
        # these are the path upto the file name.  If we're saving to a directory
        # then the path is the complete location.
        if self.PROJECTS_ARE_FILES:
            path, filename = os.path.split(loc)
        else:
            path = loc
        if len(path) > 0:
            f = File(path)
            if f.is_file:
                f.delete()
            if not f.exists:
                f.create_folders()

        # Save this project in a manner that derived classes can modify.
        self._save(loc)

        # If the save succeeds (no exceptions were raised), then update the
        # location of the project and clear the dirty flag.
        self.location = loc
        self.dirty = False

        return


    def start(self):
        """
        Notify this project that it is now the 'current' project.

        This call should only be made by the project plugin framework!

        This call *could* happen multiple times to a project, but
        only if interwoven with paired calls to the 'stop' method.

        Derived classes should override this, and chain the base
        implementation, if they need to do anything when a project
        becomes current.

        """

        logger.debug('Project [%s] started', self)

        # Ensure we start with an empty set of editors
        self._editors = {}


    def stop(self):
        """
        Called only by the project plugin framework to notify this
        project it is no longer the current project.

        This call *could* happen multiple times to a project, but
        only if interwoven with paired calls to the 'start' method.

        Derived classes should override this, and chain the base
        implementation, if they need to do anything when a project
        stops being current.

        """

        # Close all of the editors displaying our resources
        self._close_all_editors()

        logger.debug('Project [%s] stopped', self)


    #### protected interface #################################################

    def _close_all_editors(self):
        """
        Called to close all editors associated with this project.

        """

        # NOTE: The close() method on the editor will call back to remove
        # itself from our set of registered editors.  (This assumes the
        # editor is derived from ProjectEditor.)
        for editor in self._editors.values():
            logger.debug('Project requesting close of ProjectEditor [%s]',
                editor)
            editor.close()


    def _close_resource_editors(self, resource):
        """
        Close any editors associated with the specified resource(s).

        The passed value may be a single resource or a list of resources.  The
        resources should be parts of this project but no error is generated if
        they are not, nor if they are not currently associated with an editor.

        """

        # Ensure we're dealing with a list of resources.
        if not isinstance(resource, list):
            resource = [resource]

        # Close any editors associated with the resources
        for r in resource:
            editor = self._editors.get(r, None)
            if editor is not None:
                logger.debug('Requesting close of ProjectEditor [%s] from ' + \
                    'Project [%s]', editor, self)
                editor.close()

        return


    def _contains_resource(self, resource):
        """
        Called to determine if this project contains the specified
        resource.

        Note that if this project can be represented as a hierarchy, the
        resource may not be a top level object in that hierarchy!

        Derived classes must implement this!

        """

        return False


    def _get_name(self):
        """
        Returns the current name for this project.

        The name is always the last part of the path that is the project's
        location.  If we have no location, then a default name is returned.

        """

        # Prefer to use the cached version of the name
        if self._name is not None and len(self._name) > 0:
            result = self._name

        # Use (and cache) a name from our current location
        else:
            # Strip any trailing path separator off the current location so
            # that we use the last directory name if our location is a
            # directory
            location = self.location.rstrip(os.path.sep)

            # The project name is then the basename of the location
            self._name = os.path.basename(location)
            result = self._name

        return result


    def _load(cls, location):
        """
        Load a project from the specified location.

        This method exists purely to allow derived classes to have an easy way
        to override or extend the loading of a project.  The default behavior
        is to load the project from a file using the unpickle mechanism.

        The caller is notified of loading errors by raised exceptions.

        """

        # Allow derived classes to determine the actual pickle file given the
        # requested source location.
        filename = cls.get_pickle_filename(location)
        logger.debug('Loading Project of class [%s] from [%s]', cls,
            filename)

        # Try to unpickle the project while making sure to close any file we
        # opened.
        fh = None
        try:
            fh = open(filename, 'rb')
            pickle_package = cls.get_pickle_package()
            project = pickle_package.load(fh)

            # Allow derived classes to customize behavior after unpickling
            # is complete.
            project._load_hook(location)

            logger.debug('Loaded Project [%s] from location [%s]', project,
                filename)

        # Ensure any opened file is closed
        finally:
            if fh:
                try:
                    fh.close()
                except:
                    logger.exception('Unable to close project file [%s]',
                        filename)

        return project
    _load = classmethod(_load)


    def _load_hook(self, location):
        """
        Finish loading of a project.

        This method exists purely to allow derived classes to customize the
        steps that finish the loading of a project.

        Note that the project's internal location value does not yet reflect
        the location the project was loaded from, nor do we guarantee the dirty
        flag isn't set.  (Doing the right thing to both of these is done by the
        framework after this method!)

        """

        pass


    def _location_default(self):
        """
        Generates the default value for our location trait.

        """

        return self.get_default_project_location(self.application)


    def _make_location_unique(cls, location):
        """
        Return a location, based off the specified location, that does
        not already exist on the filesystem.

        """

        result = location
        counter = 1
        while os.path.exists(result):
            result = cls._unique_name_format % (location, counter)
            counter+=1

        return result
    _make_location_unique = classmethod(_make_location_unique)


    def _save(self, location):
        """
        Save this project to the specified location.

        This method exists purely to allow derived classes to have an easy way
        to override or extend the saving of a project.  The default behavior is
        to save this project to a file using the pickle mechanism.

        The caller is notified of saving errors by raised exceptions.

        """

        # Allow derived classes to determine the actual pickle file from the
        # specified location.
        filename = self.get_pickle_filename(location)
        logger.debug('Saving Project [%s] to [%s]', self, filename)

        # Pickle the object to a file while making sure to close any file we
        # open.  Note that we can't just log or ignore errors here as the
        # caller needs to know whether we succeeded or not, and could possibly
        # handle the exception if they knew what it was.
        fh = None
        try:
            # Allow derived classes to customize behavior before pickling is
            # applied.
            self._save_hook(location)

            fh = open(filename, 'wb')
            pickle_package = self.get_pickle_package()
            pickle_package.dump(self, fh, 1)

            logger.debug('Saved Project [%s] to [%s]', self, filename)
        finally:
            try:
                if fh is not None:
                    fh.close()
            except:
                logger.exception('Unable to close project pickle file [%s]',
                    filename)

        return


    def _save_hook(self, location):
        """
        Start saving a project.

        This method exists purely to allow derived classes to customize the
        steps that initiate the saving of a project.

        Note that the project's internal location value does not reflect the
        target location for the save.

        """

        pass


    #### trait handlers ######################################################

    def _location_changed(self, old, new):
        """
        Called whenever the project's location changes.

        """

        logger.debug('Location changed from [%s] to [%s] for Project [%s]',
            old, new, self)

        # Invalidate any cached project name
        old_name = self._name
        self._name = ''
        self.trait_property_changed('name', old_name, self.name)

        # Indicate this project is now dirty
        self.dirty = True


#### EOF #####################################################################



import unittest

try:
    import IPython  # noqa
except ImportError:
    from nose.plugins.skip import SkipTest
    raise SkipTest('IPython not available')

from ipykernel.kernelapp import IPKernelApp

from envisage.api import Application, Plugin
from envisage.core_plugin import CorePlugin
from envisage.plugins.ipython_kernel.internal_ipkernel import InternalIPKernel
from envisage.plugins.ipython_kernel.ipython_kernel_plugin import (
    IPYTHON_KERNEL_PROTOCOL, IPYTHON_NAMESPACE, IPythonKernelPlugin)
from traits.api import List


class TestIPythonKernelPlugin(unittest.TestCase):

    def tearDown(self):
        IPKernelApp.clear_instance()

    def test_kernel_service(self):
        # See that we can get the IPython kernel service when the plugin is
        # there.
        plugins = [CorePlugin(), IPythonKernelPlugin()]
        app = Application(plugins=plugins, id='test')

        # Starting the app starts the kernel plugin. The IPython kernel
        # service should now be available.
        app.start()
        kernel = app.get_service(IPYTHON_KERNEL_PROTOCOL)
        self.assertIsNotNone(kernel)
        self.assertIsInstance(kernel, InternalIPKernel)

        # Initialize the kernel. Normally, the application does it when
        # it starts.
        kernel.init_ipkernel(gui_backend=None)

        # Stopping the application should shut the kernel down.
        app.stop()
        self.assertIsNone(kernel.ipkernel)

    def test_kernel_namespace_extension_point(self):
        class NamespacePlugin(Plugin):
            kernel_namespace = List(contributes_to=IPYTHON_NAMESPACE)

            def _kernel_namespace_default(self):
                return [('y', 'hi')]

        plugins = [CorePlugin(), IPythonKernelPlugin(), NamespacePlugin()]
        app = Application(plugins=plugins, id='test')

        app.start()
        try:
            kernel = app.get_service(IPYTHON_KERNEL_PROTOCOL)
            kernel.init_ipkernel(gui_backend=None)
            self.assertIn('y', kernel.namespace)
            self.assertEqual(kernel.namespace['y'], 'hi')
        finally:
            app.stop()



""" The interface for objects using the 'ExtensionPoint' trait type. """


# Enthought library imports.
from traits.api import Instance, Interface

# Local imports.
from .i_extension_registry import IExtensionRegistry


class IExtensionPointUser(Interface):
    """ The interface for objects using the 'ExtensionPoint' trait type. """

    # The extension registry that the object's extension points are stored in.
    extension_registry = Instance(IExtensionRegistry)

#### EOF ######################################################################



""" Charm - the model for a simple Python IDE. """


# Enthought library imports.
from envisage.developer.api import CodeBrowser, Module
from traits.api import Event, HasTraits, Instance, Str


class Charm(HasTraits):
    """ Charm - the model for a simple Python IDE. """

    #### 'Charm' interface ####################################################

    # The code browser.
    code_browser = Instance(CodeBrowser)

    # The filename of the code database.
    filename = Str('code_database.pickle')

    # The current module.
    module = Instance(Module)

    #### Events ####

    # Fired when a module is about to be parsed.
    parsing_module = Event

    # Fired when a module has been parsed.
    parsed_module = Event

    ###########################################################################
    # 'Charm' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _code_browser_default(self):
        """ Trait initializer. """

        code_browser = CodeBrowser(filename=self.filename)
        self._setup_code_browser(code_browser)

        return code_browser

    #### Methods ##############################################################

    def load(self):
        """ Loads the code database. """

        self.code_browser.load()

        return

    def save(self):
        """ Saves the code database. """

        self.code_browser.save()

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _setup_code_browser(self, code_browser):
        """ Adds trait change listeners to a code browser. """

        code_browser.on_trait_change(self._on_parsing_module, 'parsing_module')
        code_browser.on_trait_change(self._on_parsed_module, 'parsed_module')

        return

    def _tear_down_code_browser(self, code_browser):
        """ Removes trait change listeners from a code browser. """

        code_browser.on_trait_change(
            self._on_parsing_module, 'parsing_module', remove=True
        )

        code_browser.on_trait_change(
            self._on_parsed_module, 'parsed_module', remove=True
        )

        return

    #### Trait change handlers ################################################

    def _code_browser_changed(self, old, new):
        """ Static trait change handler. """

        if old is not None:
            self._tear_down_code_browser(old)

        if new is not None:
            self._setup_code_browser(new)

        return

    def _filename_changed(self, old, new):
        """ Static trait change handler. """

        self.code_browser.filename = new

        return

    def _on_parsing_module(self, event):
        """ Dysnamic trait change handler. """

        self.parsing_module = event

        return

    def _on_parsed_module(self, event):
        """ Dysnamic trait change handler. """

        self.parsed_module = event

        return

#### EOF ######################################################################



""" An action that exits the workbench. """


# Enthought library imports.
from pyface.api import ImageResource
from pyface.action.api import Action


class ExitAction(Action):
    """ An action that exits the workbench. """

    #### 'Action' interface ###################################################

    # A longer description of the action.
    description = 'Exit the application'

    # The action's image (displayed on tool bar tools etc).
    image = ImageResource('exit')

    # The action's name (displayed on menus/tool bar tools etc).
    name = 'Exit'

    # A short description of the action used for tooltip text etc.
    tooltip = 'Exit the application'

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Perform the action. """

        self.window.application.exit()

        return

#### EOF ######################################################################



""" The application interface. """


# Enthought library imports.
from apptools.preferences.api import IPreferences
from traits.api import Event, Instance, Str, VetoableEvent

# Local imports.
from .i_extension_registry import IExtensionRegistry
from .i_import_manager import IImportManager
from .i_plugin_manager import IPluginManager
from .i_service_registry import IServiceRegistry
from .application_event import ApplicationEvent


class IApplication(
        IExtensionRegistry, IImportManager, IPluginManager, IServiceRegistry
    ):
    """ The application interface. """

    # The application's globally unique identifier.
    id = Str

    # The name of a directory (created for you) to which the application can
    # read and write non-user accessible data, i.e. configuration information,
    # preferences, etc.
    home = Str

    # The name of a directory (created for you upon access) to which the
    # application can read and write user-accessible data, e.g. projects created
    # by the user.
    user_data = Str

    # The root preferences node.
    preferences = Instance(IPreferences)

    #### Events ####

    # Fired when the application is starting. This is the first thing that
    # happens when the 'start' method is called.
    starting = VetoableEvent(ApplicationEvent)

    # Fired when all plugins have been started.
    started = Event(ApplicationEvent)

    # Fired when the plugin manager is stopping. This is the first thing that
    # happens when the 'stop' method is called.
    stopping = VetoableEvent(ApplicationEvent)

    # Fired when all plugins have been stopped.
    stopped = Event(ApplicationEvent)

    def run(self):
        """ Run the application.

        The same as::

          if application.start():
              application.stop()

        """

#### EOF ######################################################################



""" Classes used to represent functions and methods. """

# Standard library imports.
import ast

# Enthought library imports.
from traits.api import Any, HasTraits, Instance, Int, Str

# Local imports.
from .namespace import Namespace


class Function(Namespace):
    """ A function. """

    #### 'Function' interface #################################################

    # The namespace that the function is defined in.
    namespace = Instance(Namespace)

    # The line number in the module at which the function appears.
    lineno = Int

    # The name of the function.
    name = Str

    # The function's doc-string (None if there is no doc string, a string if
    # there is).
    doc = Any

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __str__(self):
        """ Returns an informal string representation of the object. """

        return 'Function %s at %d' % (self.name, self.lineno)


class FunctionFactory(HasTraits):
    """ A factory for classes. """

    ###########################################################################
    # 'FunctionFactory' interface.
    ###########################################################################

    def from_ast(self, namespace, node):
        """ Creates a class from an AST node. """

        # Create a new function.
        function = Function(
            namespace = namespace,
            lineno    = node.lineno,
            name      = node.name,
            doc       = ast.get_docstring(node, clean=False)
        )

        return function

#### EOF ######################################################################






""" A resource protocol for HTTP documents. """

# Enthought library imports.
from traits.api import HasTraits, provides

# Local imports.
from .i_resource_protocol import IResourceProtocol
from .no_such_resource_error import NoSuchResourceError


@provides(IResourceProtocol)
class HTTPResourceProtocol(HasTraits):
    """ A resource protocol for HTTP documents. """

    ###########################################################################
    # 'IResourceProtocol' interface.
    ###########################################################################

    def file(self, address):
        """ Return a readable file-like object for the specified address. """

        # Do the import here 'cos I'm not sure how much this will actually
        # be used.
        from .._compat import urlopen, HTTPError

        try:
            f = urlopen('http://' + address)

        except HTTPError:
            raise NoSuchResourceError('http:://' + address)

        return f

#### EOF ######################################################################



""" A UI for browsing a plugin. """


# Enthought library imports.
from envisage.api import ExtensionPoint, IPlugin
from traits.api import Delegate, HasTraits, Instance, List, Property
from traits.api import Code, Str
from traitsui.api import Item, TableEditor, View, VGroup
from traitsui.table_column import ObjectColumn # fixme: non-api!

class ExtensionPointModel(Hastraits):
    """ A model for browsing an extension point. """

    # The plugin that offered the extension point.
    plugin = Instance(IPlugin)

    # The extension point.
    extension_point = Instance(IExtensionPoint)

    #### 'ExtensionPointModel' interface ######################################



class ExtensionModel(Hastraits):
    """ A model for browsing a contribution to an extension point. """

    # The plugin that made the contribution.
    plugin = Instance(IPlugin)

    #### 'ContributionModel' interface ########################################

    # The Id of the extension point that the contribution is for.
    extension_point_id = Str

    # The contributions to the extension point.
    contributions = List

    ###########################################################################
    # 'ApplicationModel' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _contributions_default(self):
        """ Trait initializer. """

        return plugin.application.get_extensions(self.extension_point_id)


class PluginModel(HasTraits):
    """ A model for browsing a plugin.  """

    # The plugin that we are a model for.
    plugin = Instance(IPlugin)

    #### 'PluginModel' interface ##############################################

    # Models for each of the plugin's extension points.
    extension_point_models = List(ExtensionPointModel)

    # Models for each of the plugin's contributions to extension points.
    extension_models = List(ExtensionModel)

    ###########################################################################
    # 'ApplicationModel' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _extension_models_default(self):
        """ Trait initializer. """

        extension_models = [
            ExtensionModel(
                plugin             = plugin,
                extension_point_id = extension_point.id
            )

            for extension_point in plugin.get_extension_points()
        ]

        return extension_models

    def _extension_point_models_default(self):
        """ Trait initializer. """

        extension_point_models = [
            ExtensionPointModel(
                plugin          = plugin,
                extension_point = extension_point
            )

            for extension_point in plugin.get_extension_points()
        ]

        return extension_point_models


class ApplicationModel(HasTraits):
    """ A model for browsing an application. """

    # The application that we are a model for.
    application = Instance(IApplication)

    #### 'ApplicationModel' interface #########################################

    # Models for each of the application's plugins.
    plugin_models = List(PluginModel)

    ###########################################################################
    # 'ApplicationModel' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _plugin_models_default(self):
        """ Trait initializer. """

        return [PluginModel(plugin=plugin) for plugin in self.application]



extension_point_table_editor = TableEditor(
    columns = [
        ObjectColumn(name='id'),
        #ObjectColumn(name='desc')
    ],

#   selected = 'extension_point_selected',
    editable  = True,
    edit_view = View(Item(name='desc', show_label=False), style='custom')
)

plugin_browser_view = View(
    VGroup(
        Item(name='id'),
        Item(name='name'),
        label='General'
    ),

    VGroup(
        Item(
            name       = 'extension_points',
            editor     = extension_point_table_editor,
            show_label = False
        ),

        label='Extension Points',
    ),

    width  = .8,
    height = .6
)


class ExtensionPointBrowser(HasTraits):
    """ The model used to view extension points.

    This browser is required because 'ExatenionPoint' instances are trait
    *types* and therefore do not have traits themselves and so this class is
    really just a 'traitified' wrapper.

    """

    #### 'ExtensionPointBrowser' interface ####################################

    # The extension point that we are browsing.
    extension_point = Instance(ExtensionPoint)

    # The extension point's globally unique Id.
    id = Str

    # The extension point's description.
    desc = Code#Str

    ###########################################################################
    # 'ExtensionPoint' browser interface.
    ###########################################################################

    def _id_default(self):
        """ Trait initializer. """

        return self.extension_point.id

    def _desc_default(self):
        """ Trait initializer. """

        desc = self.extension_point.desc.strip()

        lines = [line.replace('    ', '', 2) for line in desc.splitlines()]

        return '\n'.join(lines)


# Convenience trait to delegate an attribute to a plugin.
DelegatedToPlugin = Delegate('plugin', modify=True)


class PluginBrowser(HasTraits):
    """ A plugin browser.

    Actually, this class exists just because to use a trait editor we have
    to have a trait to edit!

    """

    #### 'IPlugin' interface ##################################################

    # The plugins Id.
    id = DelegatedToPlugin

    # The plugin name.
    name = DelegatedToPlugin

    #### 'PluginBrowser' interface ############################################

    # The extension points offered by the plugin.
    extension_points = Property(List)

    # The plugin that we are browsing.
    plugin = Instance(IPlugin)

    # The default traits UI view.
    traits_view = plugin_browser_view

    ###########################################################################
    # 'PluginBrowser' interface.
    ###########################################################################

    def _get_extension_points(self):
        """ Property getter. """

        extension_points = [
            ExtensionPointBrowser(extension_point=extension_point)

            for extension_point in self.plugin.extension_points
        ]

        return extension_points


def browse_plugin(plugin):
    """ Browse a plugin. """

    import inspect

    if inspect.isclass(plugin):
        plugin = plugin()

    plugin_browser = PluginBrowser(plugin=plugin)
    plugin_browser.configure_traits(view=plugin_browser_view)

    return

#### EOF ######################################################################



""" The developer plugin. """


# Enthought library imports.
from envisage.api import Plugin, ServiceOffer
from traits.api import List


# The package that this module is in.
PKG = '.'.join(__name__.split('.')[:-1])


class DeveloperPlugin(Plugin):
    """ The developer plugin.

    This plugin contains the non-GUI part of the tools that (hopefully) help
    developers to inspect and debug a running Envisage workbench application.

    """

    # The Ids of the extension points that this plugin contributes to.
    SERVICE_OFFERS = 'envisage.service_offers'

    #### 'IPlugin' interface ##################################################

    # The plugin's globally unique identifier.
    id = 'envisage.developer'

    # The plugin's name (suitable for displaying to the user).
    name = 'Developer'

    #### Extension points offered by this plugin ##############################

    # None.

    #### Contributions to extension points made by this plugin ################

    service_offers = List(contributes_to=SERVICE_OFFERS)

    def _service_offers_default(self):
        """ Trait initializer. """

        code_browser_class = PKG + '.code_browser.api.CodeBrowser'

        code_browser_service_offer = ServiceOffer(
            protocol = code_browser_class,
            factory  = code_browser_class,
            scope    = 'application'
        )

        return [code_browser_service_offer]

#### EOF ######################################################################



""" A view containing the contents of a Python shell namespace. """

import types

# Enthought library imports.

from envisage.plugins.python_shell.api import IPythonShell
from envisage.plugins.python_shell.view.python_shell_view import PythonShellView

from pyface.workbench.api import View

from traits.api import HasTraits, Str, Property, List, Instance, \
        DelegatesTo, cached_property

from traitsui.api import Item, TableEditor, VGroup
from traitsui.api import View as TraitsView
from traitsui.table_column import ObjectColumn
from traitsui.table_filter import RuleTableFilter
from traitsui.table_filter import MenuFilterTemplate
from traitsui.table_filter import EvalFilterTemplate
from traitsui.table_filter import RuleFilterTemplate


# Table editor definition:
filters = [EvalFilterTemplate, MenuFilterTemplate, RuleFilterTemplate]

table_editor = TableEditor(
    columns     = [
        ObjectColumn(name='name'),
        ObjectColumn(name='type'),
        ObjectColumn(name='module'),
    ],
    editable    = False,
    deletable   = False,
    sortable    = True,
    sort_model  = False,
    filters     = filters,
    search      = RuleTableFilter(),
)


def type_to_str(obj):
    """
    Make a string out `obj`'s type robustly.
    """
    typ = type(obj)
    if typ.__name__ == 'vtkobject' or typ is types.InstanceType:
        typ = obj.__class__
    if type.__module__ == '__builtin__':
        # Make things like int and str easier to read.
        return typ.__name__
    else:
        name = '%s.%s' % (typ.__module__, typ.__name__)
        return name


def module_to_str(obj):
    """
    Return the string representation of `obj`s __module__ attribute, or an
    empty string if there is no such attribute.
    """
    if hasattr(obj, '__module__'):
        return str(obj.__module__)
    else:
        return ''


class NamespaceView(View):
    """ A view containing the contents of the Python shell namespace. """

    #### 'IView' interface ####################################################

    # The part's globally unique identifier.
    id = 'enthought.plugins.python_shell.view.namespace_view'

    # The view's name.
    name = 'Namespace'

    # The default position of the view relative to the item specified in the
    # 'relative_to' trait.
    position = 'left'

    #### 'NamespaceView' interface ############################################

    # The bindings in the namespace.  This is a list of HasTraits objects with
    # 'name', 'type' and 'module' string attributes.
    bindings = Property(List, depends_on=['namespace'])

    shell_view = Instance(PythonShellView)

    namespace = DelegatesTo('shell_view')


    # The default traits UI view.
    traits_view = TraitsView(
        VGroup(
            Item(
                'bindings',
                id     = 'table',
                editor = table_editor,
                springy = True,
                resizable = True,
            ),

            show_border = True,
            show_labels = False
        ),
        resizable = True,
    )

    ###########################################################################
    # 'View' interface.
    ###########################################################################

    def create_control(self, parent):
        """ Creates the toolkit-specific control that represents the view.

        'parent' is the toolkit-specific control that is the view's parent.

        """

        self.ui = self.edit_traits(parent=parent, kind='subpanel')

        self.shell_view = self.window.application.get_service(IPythonShell)
        # 'shell_view' is an instance of the class PythonShellView from the module
        # envisage.plugins.python_shell.view.python_shell_view.

        return self.ui.control

    ###########################################################################
    # 'NamespaceView' interface.
    ###########################################################################

    #### Properties ###########################################################

    @cached_property
    def _get_bindings(self):
        """ Property getter. """

        if self.shell_view is None:
            return []

        class item(HasTraits):
            name = Str
            type = Str
            module = Str

        data = [item(name=name, type=type_to_str(value), module=module_to_str(value))
                    for name, value in self.shell_view.namespace.items()]

        return data





from traits.api import Callable, HasTraits, List, Str



class VersionInfo(HasTraits):
    """ Represents the information about a particular version of an
    application.
    """

    # The version string that this
    version = Str()

    # Customer-facing notes about this version.  Typically this is an
    # HTML document containing the changelog between this version and
    # the previous version
    notes = Str()

    # The location of where to obtain this version.  Typically this will
    # be an HTTP URL, but this can be a URI for a local or LAN item, or
    # it can be a
    location = Str()

    # A function that takes a string (self.version) and returns something
    # that can be used to compare against the version-parsed version of
    # another VersionInfo object.
    version_parser = Callable()

    def __cmp__(self, other):
        """ Allows for comparing two VersionInfo objects so they can
        be presented in version-sorted order.  This is where we parse
        and interpretation of the **version** string attribute.
        """
        # TODO: Do something more intelligent here
        if self.version_parser is not None:
            self_ver = self.version_parser(self.version)
        else:
            self_ver = self.version
        if other.version_parser is not None:
            other_ver = other.version_parser(other.version)
        else:
            other_ver = other.version
        return self_ver < other_ver

class UpdateInfo(HasTraits):
    """ Encapsulates the information about the available update or
    updates.  An update can consist of multiple versions, with each
    version containing its own information and download URL.
    """

    updates = List(VersionInfo)

    pass







# Standard library imports
import logging
import os
import select
import socket
import sys
from threading import Thread

# ETS imports
from traits.api import HasTraits, Int, Str, Bool, Instance, List, \
     Tuple, Enum
from envisage.plugins import remote_editor

# Local imports
from .server import Server
from .util import accept_no_intr, get_server_port, receive, send_port, \
    spawn_independent, MESSAGE_SEP

logger = logging.getLogger(__name__)


class ClientThread(Thread):
    """ A thread for listening for commands from the server.
    """

    def __init__(self, client):
        Thread.__init__(self)
        self.client = client
        self._finished = False

    def run(self):
        # Get the server port, spawning it if necessary
        server_port = get_server_port()
        if server_port == -1 or not Server.ping(server_port, timeout=5):
            if len(self.client.server_prefs):
                # Spawn the server
                logger.info("Client spawning Server...")
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.settimeout(5)
                sock.bind(('localhost', 0))
                sock.listen(1)
                port = sock.getsockname()[1]
                args = self.client.server_prefs + ( port, )
                code = "from envisage.plugins.remote_editor.communication." \
                    "server import main; main(r'%s', '%s', %i)" % args
                spawn_independent([sys.executable, '-c', code])

                # Await a reponse from the server
                try:
                    server, address = accept_no_intr(sock)
                    try:
                        command, arguments = receive(server)
                        if command == "__port__":
                            self.client._server_port = int(arguments)
                        else:
                            raise socket.error
                    finally:
                        # Use try...except to handle timeouts
                        try:
                            server.shutdown(socket.SHUT_RD)
                        except:
                            pass
                except socket.error as e:
                    logger.error(repr(e))
                    logger.error("Client spawned a non-responsive Server! " \
                                     "Unregistering...")
                    self.client.error = True
                    self.client.unregister()
                    return
                finally:
                    sock.close()

            else:
                logger.error("Client could not contact the Server and no " \
                                 "spawn command is defined. Unregistering...")
                self.client.error = True
                self.client.unregister()
                return
        else:
            self.client._server_port = server_port

        # Create the socket that will receive commands from the server
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('localhost', 0))
        sock.listen(1)
        self.client._port = sock.getsockname()[1]

        # Register with the server
        port = str(self.client._port)
        arguments = MESSAGE_SEP.join((port, self.client.self_type,
                                      self.client.other_type))
        self.client.error = not send_port(self.client._server_port, 'register',
                                          arguments)
        self.client.registered = True

        # Send queued commands (these can only exist if we spawned the server)
        for command, args in self.client._queue:
            arguments = MESSAGE_SEP.join((port, command, args))
            self.client.error = not send_port(self.client._server_port, 'send',
                                              arguments)
        self.client._queue = []

        # Start the loop to listen for commands from the Server
        logger.info("Client listening on port %i..." % self.client._port)
        try:
            while not self._finished:
                server, address = accept_no_intr(sock)

                # Reject non-local connections
                if address[0] != '127.0.0.1':
                    msg = "Client on port %s received connection from a " \
                        "non-local party (port %s). Ignoring..."
                    logger.warning(msg % (port, address[0]))
                    continue

                # Receive the command from the server
                self.client.error = False
                command, arguments = receive(server)
                msg = r"Client on port %s received: %s %s"
                logger.debug(msg, port, command, arguments)

                # Handle special commands from the server
                if command == "__orphaned__":
                    self.client.orphaned = bool(int(arguments))

                elif command == "__error__":
                    error_status = arguments[0]
                    error_message = ''
                    if len(arguments) > 0:
                        error_message = arguments[1:]
                    logger.warning("Error status received from the server: " \
                                       "%s\n%s" % (error_status, error_message))
                    self.client.error = bool(int(error_status))

                # Handle other commands through Client interface
                else:
                    if self.client.ui_dispatch == 'off':
                        self.client.handle_command(command, arguments)
                    else:
                        if self.client.ui_dispatch == 'auto':
                            from pyface.gui import GUI
                        else:
                            exec('from pyface.ui.%s.gui import GUI' %
                                 self.client.ui_dispatch)
                        GUI.invoke_later(self.client.handle_command,
                                         command, arguments)
        finally:
            self.client.unregister()

    def stop(self):
        self._finished = True


class Client(HasTraits):
    """ An object that communicates with another object through a Server.
    """

    # The preferences file path and node path to use for spawning a Server. If
    # this is not specified it will not be possible for this Client to spawn
    # the server.
    server_prefs = Tuple((os.path.join(remote_editor.__path__[0],
                                       "preferences.ini"),
                          "enthought.remote_editor"),
                         Str, Str)

    # The type of this object and the type of the desired object, respectively
    self_type = Str
    other_type = Str

    # Specifies how 'handle_command' should be called. If 'auto', use the
    # dispatch method appropriate for the toolkit Traits is using. Failure to
    # set this variable as appropriate will likely result in crashes.
    ui_dispatch = Enum('off', 'auto', 'wx', 'qt4')

    # Whether this client has been registered with the Server. Note that this is
    # *not* set after the 'register' method is called--it is set when the Server
    # actually receives the register command, which happens asynchronously.
    registered = Bool(False)

    # The client's orphaned status
    orphaned = Bool(True)

    # The client's error state. This is set to True after a failure to
    # communicate with the server or a failure by the server to communicate with
    # this object's counterpart.
    error = Bool(False)

    # Protected traits
    _port = Int
    _server_port = Int
    _communication_thread = Instance(ClientThread)
    _queue = List(Tuple(Str, Str))

    def register(self):
        """ Inform the server that this Client is available to receive
            commands.
        """
        if self._communication_thread is not None:
            raise RuntimeError("'register' has already been called on Client!")

        self._communication_thread = ClientThread(self)
        self._communication_thread.setDaemon(True)
        self._communication_thread.start()

    def unregister(self):
        """ Inform the server that this Client is no longer available to receive
            commands. Note that it is poor etiquette not to communicate when a
            Client is becoming unavailable. Calling 'unregister' when a Client
            is not registered has no effect.
        """
        if self._communication_thread is not None:
            self._communication_thread.stop()
            self._communication_thread = None
            send_port(self._server_port, 'unregister', str(self._port))
            self._port = 0
            self._server_port = 0
            self.registered = False
            self.orphaned = True

    def send_command(self, command, arguments=''):
        """ Send a command to the server which is to be passed to an object
            of the appropriate type.
        """
        if self._communication_thread is None:
            raise RuntimeError("Client is not registered. Cannot send command.")

        msg = r"Client on port %i sending: %s %s"
        logger.debug(msg, self._port, command, arguments)

        args = MESSAGE_SEP.join((str(self._port), command, arguments))
        if self.registered:
            self.error = not send_port(self._server_port, 'send', args)
        else:
            self._queue.append((command, arguments))

    def handle_command(self, command, arguments):
        """ This function should take a command string and an arguments string
            and do something with them. It should return True if the command
            given was understood; otherwise, False.
        """
        raise NotImplementedError()







# Standard library imports
import os, sys
import logging
import socket

# ETS imports
from apptools.preferences.api import Preferences
from traits.api import HasTraits, HasStrictTraits, Int, Str, List, \
     Dict, Tuple, Instance

# Local imports
from .util import accept_no_intr, receive, send_port, spawn_independent, \
    MESSAGE_SEP, LOCK_PATH, LOG_PATH

logger = logging.getLogger("communication")
logger.setLevel(logging.DEBUG)
logging.basicConfig(filename=LOG_PATH)


class PortInfo(HasStrictTraits):
    """ Object to store information on how a port is being used.
    """
    port = Int
    type = Str
    other_type = Str

    def __str__(self):
        return "Port: %i, Type: %s, Other type: %s" % \
            (self.port, self.type, self.other_type)


class Server(HasTraits):
    """ A socket protocal that facilates two objects communicating with each
        other.

        The only instance methods that should be called in ordinary use are
        'init' and 'main'. Communication should be done through the methods on
        'Client' objects.
    """

    spawn_commands = Dict(Str, Str)

    _port = Int
    _sock = Instance(socket.socket)
    _port_map = Dict(Int, Instance(PortInfo))
    _orphans = List(Instance(PortInfo))
    _pairs = Dict(Instance(PortInfo), Instance(PortInfo))

    # Commands that have been queued for spawned process
    # desired_type -> list of (command, arguments)
    _queue = Dict(Str, List(Tuple(Str, Str)))

    def init(self, pref_path='', pref_node=''):
        """ Read a configuration file and attempt to bind the server to the
            specified port.
        """
        # Bind to port and write port to lock file
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.bind(('localhost', 0))
        self._port = self._sock.getsockname()[1]
        f = open(LOCK_PATH, 'w')
        f.write(str(self._port))
        f.close()

        # Read configuration file
        if not pref_path:
            return
        if os.path.exists(pref_path):
            prefs = Preferences(filename=pref_path)
            if prefs.node_exists(pref_node):
                spawn_node = prefs.node(pref_node)
                for key in spawn_node.keys():
                    cmd = spawn_node.get(key).strip()
                    if cmd.startswith('python '):
                        cmd = cmd.replace('python', sys.executable, 1)
                    self.spawn_commands[key] = cmd
            else:
                msg = "Server could not locate preference node '%s.'"
                logger.error(msg % pref_node)
        else:
            msg = "Server given non-existent preference path '%s'."
            logger.error(msg % pref_path)

    def main(self, port=0):
        """ Starts the server mainloop. If 'port' is specified, the assumption
            is that this Server was spawned from an object on said port. The
            Server will inform the Client that it has been created, and if fails
            to contact the client, this function will return.
        """
        try:
            # Start listening *before* we (potentially) inform the spawner that
            # we are working. This means that if the spawner tries to register,
            # we will be ready.
            logger.info("Server listening on port %i..." % self._port)
            self._sock.listen(5)
            self._sock.settimeout(300)

            # If necessary, inform the launcher that we have initialized
            # correctly by telling it our port
            if port:
                if not send_port(port, "__port__", str(self._port), timeout=5):
                    msg = "Server could not contact spawner. Shutting down..."
                    logger.warning(msg)
                    return

            # Start the mainloop
            while True:
                try:
                    client, address = accept_no_intr(self._sock)
                except socket.timeout:
                    # Every 5 minutes of inactivity, we trigger a garbage
                    # collection. We do this to make sure the server doesn't
                    # stay on, with dead process as zombies.
                    self._gc()
                    continue
                try:
                    if address[0] != '127.0.0.1':
                        msg = "Server received connection from a non-local " \
                            "party (port %s). Ignoring..."
                        logger.warning(msg, address[0])
                        continue
                    command, arguments = receive(client)

                    logger.debug("Server received: %s %s", command, arguments)

                    if command == "send":
                        port, command, arguments = arguments.split(MESSAGE_SEP)
                        self._send_from(int(port), command, arguments)
                    elif command == "register":
                        port, type, other_type = arguments.split(MESSAGE_SEP)
                        self._register(int(port), type, other_type)
                    elif command == "unregister":
                        self._unregister(int(arguments))
                    elif command == "ping":
                        self._send_to(int(arguments), "__status__", "1")
                    elif command == "spawn":
                        self._spawn(arguments)
                    else:
                        logger.error("Server received unknown command: %s %s",
                                     command, arguments)
                finally:
                    client.close()
        finally:
            self._sock.close()

    @staticmethod
    def ping(server_port, timeout=1, error_only=False):
        """ Returns whether the server is running on 'server_port'.

            If error_only, return False only if there was an error (ie the
            socket is closed).
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout)
        sock.bind(('localhost', 0))
        sock.listen(1)
        try:
            port = str(sock.getsockname()[1])
            send_port(server_port, "ping", port, timeout=timeout)
            try:
                server, address = accept_no_intr(sock)
                try:
                    command, arguments = receive(server)
                    return command == "__status__" and bool(int(arguments))
                finally:
                    try:
                        server.shutdown(socket.SHUT_RD)
                    except:
                        pass
            except socket.error:
                # 'sock' may have timed out, so use try..except.
                try:
                    sock.shutdown(socket.SHUT_RD)
                except:
                    pass
                return False
        except:
            try:
                sock.shutdown(socket.SHUT_RD)
            except:
                pass

    def _spawn(self, object_type):
        """ Attempt to spawn an process according the specified type. Returns
            whether this was sucessful.
        """
        try:
            command = self.spawn_commands[object_type]
        except KeyError:
            msg = "No spawn command is defined for object type '%s'." \
                % object_type
            logger.warning(msg)
            return msg

        try:
            logger.info("Server spawning Client of type '%s.'" % object_type)
            spawn_independent(command, shell=True)
        except OSError:
            msg = "Error spawning process for '%s' with command '%s'." \
                % (object_type, command)
            logger.error(msg)
            return msg

        return False

    def _match(self, port):
        """ Attempt to match a registered client on 'port' to another registered
            client of the same type.
        """
        # Try to find a compatible orphan
        info = self._port_map[port]
        for orphan in self._orphans:
            if info.other_type == orphan.type:

                # Ping the orphan to see if its alive:
                if not self._send_to(orphan.port, '__keepalive__', ''):
                    continue

                # Save information about the matching
                self._orphans.remove(orphan)
                self._pairs[info] = orphan
                self._pairs[orphan] = info

                # Dispatch orphaned status to clients
                self._send_to(port, "__orphaned__", "0")
                self._send_to(orphan.port, "__orphaned__", "0")

                # Check command queue and dispatch, if necessary
                for command, arguments in self._queue.pop(info.type, []):
                    self._send_to(port, command, arguments)

                return orphan

        # Otherwise, the object becomes an orphan
        self._orphans.append(info)
        return None

    def _register(self, port, object_type, other_type):
        """ Register a port of 'object_type' that wants to be paired with
            another object of 'other_type'. These types are simply strings.

            Calling 'register' on an already registered port has no effect.
        """
        # Only continue if this object is not already registered
        if port in self._port_map:
            return

        info = PortInfo(port=port, type=object_type, other_type=other_type)
        self._port_map[port] = info
        self._match(port)

    def _unregister(self, port):
        """ Unregister a port. Calling 'unregister' on a port that is not
            registered has no effect.
        """
        try:
            info = self._port_map.pop(port)
        except KeyError:
            return

        if info in self._pairs:
            other = self._pairs.pop(info)
            if other in self._pairs:
                self._pairs.pop(other)
            self._orphans.append(other)
            self._send_to(other.port, "__orphaned__", "1")
        else:
            self._orphans.remove(info)

        # If we have nobody registered, terminate the server.
        if len(self._port_map) == 0:
            logger.info("No registered Clients left. Server shutting down...")
            sys.exit(0)

    def _gc(self):
        """ Garbage collection of the processes. Check that all the orphaned
            processes are still responsive, and if not, unregister them.
        """
        for object_info in self._orphans:
            self._send_to(object_info.port, '__keepalive__', '')
            # _send_to will automatically unregister the port.

    def _send_from(self, port, command, arguments):
        """ Send a command from an object on the specified port.
        """
        try:
            object_info = self._port_map[port]
        except KeyError:
            msg = "Server received a 'send' command from an unregistered " \
                "object (port %s)."
            logger.warning(msg % port)
            return

        try_spawn = False
        if object_info in self._pairs:
            other = self._pairs[object_info]
            while not self._send_to(other.port, command, arguments):
                # The object will automatically be orphaned by _send_to in the
                # event of a communication failure. Try to find another match.
                if object_info in self._orphans:
                    self._orphans.remove(object_info)
                other = self._match(port)
                if other is None:
                    try_spawn = True
                    break
        else:
            try_spawn = True

        if try_spawn:
            queue = self._queue.get(object_info.other_type)
            if queue:
                queue.append((command, arguments))
            else:
                error_msg = self._spawn(object_info.other_type)
                if error_msg:
                    self._send_to(port, '__error__', '1' + error_msg)
                else:
                    self._queue[object_info.other_type] = [(command, arguments)]
                    self._match(port)

    def _send_to(self, port, command, arguments):
        """ Send a command to an object on the specified port. Returns whether
            the command was sent sucessfully.
        """
        status = send_port(port, command, arguments)
        if not status:
            msg = "Server failed to communicate with client on port %i. " \
                "Unregistering..."
            logger.warning(msg % port)
            self._unregister(port)
        return status


def main(pref_path, pref_node, *arg, **kw):
    server = Server()
    server.init(pref_path, pref_node)
    server.main(*arg, **kw)
